'''Event Log Monitor Module'''
# import re
# from os import path
import sys
import time
import json
import serial
#
# from miva.module_miva import Miva
# from module_utils import scan_serial_ports
from pyZynoUnifiedDrivers.miva.module_utils import EVENT_LOG_NUMBER_OF_EVENTS
from pyZynoUnifiedDrivers.miva.module_utils import get_line_number
from pyZynoUnifiedDrivers.miva.module_utils import _OK_KEY, _SHORT_PRESS, _RUN_KEY
from pyZynoUnifiedDrivers.miva.module_infusion import InfusionStatus

# Time Out (seconds)
_TIMEOUT = 1

# Bolus infusion rate (mL/hr)
_BOLUS_RATE = 250

# class PumpEvent(Enum):
    # '''Pump Event'''
    # NONE = 0
    # RUN_INFUSION = 1
    # INFUSION_PAUSED = 2
    # INFUSION_COMPLETED = 3
    # INFUSION_LIMIT_REACHED = 4
    # AUTHENTICATION = 5
    # BOLUS_GRANTED = 6
    # BOLUS_DENIED = 7
    # DOSE_CANCEL = 8
    # OCCLUSION = 9
    # DELAYED_END = 10
    # PUMP_LIMIT_REACHED_ALARM = 11
    # CLEAR_SHIFT_TOTAL = 12
    # BATTERY = 13
    # OPEN_CASSETTE = 14
    # TIMESTAMP = 15
    # POWER_ON = 16
    # POWER_OFF = 17
    # FIRMWARE_ERROR = 18
    # SYSTEM_ERROR = 19
    # BATTERY_FAILURE = 20
    # DEBUG = 21
    # UNATTENDED = 22


class EventLogMonitor:
    '''Event Log Monitor Class'''

    def __init__(self):
        '''__init__'''
        self._is_on = False
        self._is_paused = True
        self.event_log_index_head = '0000'
        self.timestamp_infusion_start = time.time()
        self.local_time_infusion_start = time.localtime()
        #
        self.miva = None
        #
        self.infusion = None

    def start(self, miva, infusion=None):
        '''start'''
        #
        self.miva = miva
        self._is_on = True
        self._is_paused = False
        if infusion is not None:
            self.infusion = infusion
        # print('start event log monitor...')
        try:
            event_log_index = miva.read_event_log_indices()
            previous_head = int(event_log_index['head'], 16)
        except serial.serialutil.SerialException:
            # Stop event log monitor
            self.stop()
            print('Error: pump failed to respond')
            print('Stop event log monitor...')
            return
        while self._is_on:
            try:
                if self._is_paused:
                    time.sleep(_TIMEOUT)
                    continue
                event_log_index = miva.read_event_log_indices()
                head = int(event_log_index['head'], 16)
                total_length = head - previous_head if head >= previous_head else \
                                head + EVENT_LOG_NUMBER_OF_EVENTS - previous_head
                if total_length > 0:
                    event_logs = miva.read_range_event_log(previous_head, head)
                    print('--')
                    print('total_length = [{}]'.format(total_length))
                    print('--')
                    event_logs_json = json.dumps(event_logs, indent=4)
                    print(event_logs_json)
                    print('Total Event Log Printed: [{}]'.format(len(event_logs)))
                    print('--')
                    previous_head = head
                    # check event logs for certain events
                    for each_event_log in event_logs:
                        if each_event_log.get('event_type') is not None:
                            if each_event_log['event_type'] == 'UNATTENDED_ALARM':
                                # Send [OK] key command
                                result = miva.press_key(_OK_KEY, _SHORT_PRESS)
                                if result:
                                    print('key: Ok/Enter (short)')
                            if each_event_log['event_type'] == 'RUN_INFUSION':
                                self.timestamp_infusion_start = time.time()
                                print("Infusion start at [{}]".format(time.strftime('%H:%M:%S', \
                                        time.localtime())))
                                self.local_time_infusion_start = time.localtime()
                                # Run infusion
                                if infusion is not None and infusion.protocol is not None:
                                    if infusion.status == InfusionStatus.STOPPED:
                                        infusion.run_infusion()
                                    elif infusion.status == InfusionStatus.PAUSED:
                                        infusion.pause()
                                    elif infusion.status == InfusionStatus.PAUSED_AUTO_BOLUS:
                                        infusion.pause()
                                    elif infusion.status == InfusionStatus.PAUSED_DEMAND_BOLUS:
                                        infusion.pause()
                                    elif infusion.status == InfusionStatus.PAUSED_LOADING_DOSE:
                                        infusion.pause()
                                    elif infusion.status == InfusionStatus.PAUSED_CLINICIAN_DOSE:
                                        infusion.pause()
                            if each_event_log['event_type'] == 'INFUSION_PAUSED':
                                timestamp_infusion_pause = time.time()
                                local_time_start = self.local_time_infusion_start
                                print("Infusion start at [{}]"\
                                        .format(time.strftime('%H:%M:%S', local_time_start)))
                                print("Infusion pause at [{}]"\
                                        .format(time.strftime('%H:%M:%S', time.localtime())))
                                timestamp_duration = timestamp_infusion_pause \
                                        -self.timestamp_infusion_start
                                duration = time.strftime('%H:%M:%S', \
                                        time.gmtime(timestamp_duration))
                                print('Infusion duration: [{}]'.format(duration))
                                # Pause infusion
                                if infusion is not None:
                                    print()
                                    print('line {0}: infusion.pause() [{1}]'\
                                            .format(get_line_number(), \
                                            time.strftime('%H:%M:%S', time.localtime())))
                                    infusion.pause()
                            if each_event_log['event_type'] == 'INFUSION_COMPLETED':
                                if infusion is not None:
                                    infusion.reset()
                                timestamp_infusion_complete = time.time()
                                local_time_start = self.local_time_infusion_start
                                print("line {0}: Infusion start at [{1}]"\
                                      .format(get_line_number(), \
                                              time.strftime('%H:%M:%S', local_time_start)))
                                print("line {0}: Infusion complete at [{1}]"\
                                      .format(get_line_number(), \
                                              time.strftime('%H:%M:%S', time.localtime())))
                                timestamp_duration = timestamp_infusion_complete \
                                        -self.timestamp_infusion_start
                                duration = time.strftime('%H:%M:%S', \
                                                         time.gmtime(timestamp_duration))
                                print('line {0}: Infusion duration: [{1}]'\
                                      .format(get_line_number(), duration))
                                #
                                # Send [RUN] key [SHORT] command to confirm infusion complete
                                result = miva.press_key(_RUN_KEY, _SHORT_PRESS)
                                if result:
                                    print('key: Run/Stop (short)')
                            if each_event_log['event_type'] == 'DELAY_END' and \
                                    each_event_log['sub_type'] == 'NORMAL_END':
                                if infusion is not None:
                                    infusion.trigger_delay_end()
                            if each_event_log['event_type'] == 'INFUSION_LIMIT_REACHED' and \
                                    each_event_log['sub_type'] == 'MAX_PER_HOUR':
                                if infusion is not None:
                                    infusion.trigger_mph()
                            if each_event_log['event_type'] == 'INFUSION_LIMIT_REACHED' and \
                                    each_event_log['sub_type'] == 'MAX_PER_INTERVAL':
                                if infusion is not None:
                                    infusion.trigger_mpi()
                            if each_event_log['event_type'] == 'INFUSION_LIMIT_REACHED' and \
                                    each_event_log['sub_type'] == 'CLEARED':
                                if infusion is not None:
                                    infusion.trigger_mpi_mph_clear()
                            if each_event_log['event_type'] == 'BOLUS_GRANTED' and \
                                    each_event_log['sub_type'] == 'AUTO_BOLUS':
                                if infusion is not None:
                                    infusion.trigger_ab()
                            if each_event_log['event_type'] == 'BOLUS_GRANTED' and \
                                    each_event_log['sub_type'] == 'EXTRA_DOSE':
                                if infusion is not None: 
                                    infusion.trigger_db()
                            if each_event_log['event_type'] == 'BOLUS_GRANTED' and \
                                    each_event_log['sub_type'] == 'CLINICIAN_DOSE':
                                if infusion is not None:
                                    infusion.trigger_cd()
                time.sleep(_TIMEOUT)
            except (serial.serialutil.SerialException, ValueError):
                # Stop event log monitor
                self.stop()
                print('Error: pump failed to respond')
                print('Stop event log monitor...')
                break
            except (OSError, TypeError, NameError):
                print()
                print("{0}: {1}\n".format(sys.exc_info()[0], sys.exc_info()[1]))
                self.pause()
                raise

        print('Event log monitor stopped at [{}]'\
                .format(time.strftime('%H:%M:%S', time.localtime())))
        # Stop event log monitor
        self.stop()

    def stop(self):
        '''stop'''
        # print('\nStop event log monitor...')
        self._is_paused = True
        self._is_on = False

    def is_on(self):
        '''is on'''
        return self._is_on

    def is_paused(self):
        ''''is paused'''
        is_paused = True
        if self._is_on:
            is_paused = self._is_paused
        else:
            print('Abort: Event log monitor is [OFF]')
            is_paused = None
        return is_paused

    def pause(self):
        ''''pause'''
        if self._is_on:
            self._is_paused = True
        else:
            print('Pause ELM Aborted: Event log monitor is [OFF]')

    def resume(self):
        ''''resume'''
        if self._is_on:
            self._is_paused = False
        else:
            print('Resume ELM Aborted:: Event log monitor is [OFF]')
            
# def main(argv):
#     '''main function'''
#     try:
#         current_file = path.basename(__file__)
#         if len(argv) == 1:
#             # Print the name of current script in Python
#             print('Empty argument list. ', end='')
#             print('To start, type <{}> followed by the com port. '.format(current_file), end='')
#             print('Ex: \'{} COM1\''.format(current_file))
#             serial_port_list = scan_serial_ports()
#             print('\tAvailable Serial Ports:')
#             print('\t\t\t\t', end='')
#             for each_serial_port in serial_port_list:
#                 print('{} '.format(each_serial_port), end='')
#         else:
#             serial_port = argv[1].upper()
#             if re.match(r'com\d+', serial_port.lower().rstrip(' \t\r\n\0')):
#                 miva = Miva(serial_port)
#                 event_log_monitor = EventLogMonitor()
#                 event_log_monitor.start(miva)
#             else:
#                 print('Error: invalid port \'{}\' '.format(serial_port), end='')
#                 print('To start, type \'{}\' followed by the com port. '\
#                         .format(current_file), end='')
#                 print('Ex: \'{} COM1\''.format(current_file))
#     except KeyboardInterrupt:
#         pass
#     except serial.serialutil.SerialException:
#         print("{0}: {1}\n".format(sys.exc_info()[0], sys.exc_info()[1]))

#
# if __name__ == "__main__":
#     main(sys.argv)



'''class_motor'''
import sys
import math
import queue
#import threading
from threading import Thread, Lock
import time
#
# from infusion_mode import InfusionMode

from enum import Enum

class MotorStatus(Enum):
    '''infusion mode'''
    STOPPED = 0
    RUNNING = 1
    PAUSED = 2

# the time needed for the motor to finish one rotation
_INTERVAL = 0.01
_REFRESH_TIMER = 0.1

class Motor:
    '''motor class
    mph - max per hour
    mpi - max per interval
    '''
    def __init__(self):
        self.rate = 0
        self.previous_rate = 0
        self.rate_basal = 0
        self.vinf = 0
        # auto bolus vinf for each ab interval
        self.vinf_ab = 0
        # demand bolus vinf for each demand bolus
        self.vinf_db = 0
        # clinician dose vinf for each clinician dose
        self.vinf_cd = 0
        # intermittent dose vinf for each intermittent dose interval
        self.vinf_dose = 0
        # mph - max per hour counter
        self.mph = 0
        self.mph_queue = None
        # mpi - max per interval counter
        self.mpi = 0
        self.time_passed = 0
        # time passed since auto bolus start
        self.time_passed_ab = 0
        # time passed since demand bolus start
        self.time_passed_db = 0
        # time passed since clinician dose start
        self.time_passed_cd = 0
        # time passed since intermittent dose start
        self.time_passed_id = 0
        # time passed since delay start
        self.time_passed_dl = 0
        self.status = MotorStatus.STOPPED
        # Create a mutex
        self.lock = Lock()

    def pause(self):
        '''pause motor'''
        ############################################################################
        ##                         CRITICAL SECTION START                         ##
        #                                                                          #

        # Acquire Mutex
        self.lock.acquire()

        self.status = MotorStatus.PAUSED

        # Release Mutex
        self.lock.release()

        #                                                                          #
        ##                         CRITICAL SECTION END                           ##
        ############################################################################

    def resume(self):
        '''resume motor'''
        ############################################################################
        ##                         CRITICAL SECTION START                         ##
        #                                                                          #

        # Acquire Mutex
        self.lock.acquire()

        self.status = MotorStatus.RUNNING

        # Release Mutex
        self.lock.release()

        #                                                                          #
        ##                         CRITICAL SECTION END                           ##
        ############################################################################

    def set_rate(self, motor_rate):
        '''set rate'''
        ############################################################################
        ##                         CRITICAL SECTION START                         ##
        #                                                                          #

        # Acquire Mutex
        self.lock.acquire()

        self.rate = motor_rate

        # Release Mutex
        self.lock.release()

        #                                                                          #
        ##                         CRITICAL SECTION END                           ##
        ############################################################################
    def set_previous_rate(self, motor_rate):
        '''set previous rate'''
        ############################################################################
        ##                         CRITICAL SECTION START                         ##
        #                                                                          #

        # Acquire Mutex
        self.lock.acquire()

        self.previous_rate = motor_rate

        # Release Mutex
        self.lock.release()

        #                                                                          #
        ##                         CRITICAL SECTION END                           ##
        ############################################################################
    def set_rate_basal(self, basal_rate):
        '''set rate'''
        ############################################################################
        ##                         CRITICAL SECTION START                         ##
        #                                                                          #

        # Acquire Mutex
        self.lock.acquire()

        self.rate_basal = basal_rate

        # Release Mutex
        self.lock.release()

        #                                                                          #
        ##                         CRITICAL SECTION END                           ##
        ############################################################################

    def set_vinf(self, motor_vinf):
        '''set rate'''
        ############################################################################
        ##                         CRITICAL SECTION START                         ##
        #

        # Acquire Mutex
        self.lock.acquire()

        self.vinf = motor_vinf

        # Release Mutex
        self.lock.release()

        #                                                                          #
        ##                         CRITICAL SECTION END                           ##
        ############################################################################

    def set_vinf_ab(self, vinf):
        '''set auto bolus vinf'''
        ############################################################################
        ##                         CRITICAL SECTION START                         ##
        #

        # Acquire Mutex
        self.lock.acquire()

        self.vinf_ab = vinf

        # Release Mutex
        self.lock.release()

        #                                                                          #
        ##                         CRITICAL SECTION END                           ##
        ############################################################################

    def set_vinf_db(self, vinf):
        '''set demand bolus vinf'''
        ############################################################################
        ##                         CRITICAL SECTION START                         ##
        #

        # Acquire Mutex
        self.lock.acquire()

        self.vinf_db = vinf

        # Release Mutex
        self.lock.release()

        #                                                                          #
        ##                         CRITICAL SECTION END                           ##
        ############################################################################

    def set_vinf_cd(self, vinf):
        '''set clinician dose vinf'''
        ############################################################################
        ##                         CRITICAL SECTION START                         ##
        #

        # Acquire Mutex
        self.lock.acquire()

        self.vinf_cd = vinf

        # Release Mutex
        self.lock.release()

        #                                                                          #
        ##                         CRITICAL SECTION END                           ##
        ############################################################################

    def set_vinf_dose(self, vinf):
        '''set intermittent dose vinf'''
        ############################################################################
        ##                         CRITICAL SECTION START                         ##
        #

        # Acquire Mutex
        self.lock.acquire()

        self.vinf_dose = vinf

        # Release Mutex
        self.lock.release()

        #                                                                          #
        ##                         CRITICAL SECTION END                           ##
        ############################################################################

    def set_time_passed(self, time_passed):
        '''set time passed'''
        ############################################################################
        ##                         CRITICAL SECTION START                         ##
        #

        # Acquire Mutex
        self.lock.acquire()

        self.time_passed = time_passed

        # Release Mutex
        self.lock.release()

        #                                                                          #
        ##                         CRITICAL SECTION END                           ##
        ############################################################################


    def set_time_passed_ab(self, time_passed):
        '''set time passed of the auto bolus interval'''
        ############################################################################
        ##                         CRITICAL SECTION START                         ##
        #

        # Acquire Mutex
        self.lock.acquire()

        self.time_passed_ab = time_passed

        # Release Mutex
        self.lock.release()

        #                                                                          #
        ##                         CRITICAL SECTION END                           ##
        ############################################################################

    def set_time_passed_db(self, time_passed):
        '''set time passed since demand bolus start'''
        ############################################################################
        ##                         CRITICAL SECTION START                         ##
        #

        # Acquire Mutex
        self.lock.acquire()

        self.time_passed_db = time_passed

        # Release Mutex
        self.lock.release()

        #                                                                          #
        ##                         CRITICAL SECTION END                           ##
        ############################################################################

    def set_time_passed_cd(self, time_passed):
        '''set time passed since clinician dose start'''
        ############################################################################
        ##                         CRITICAL SECTION START                         ##
        #

        # Acquire Mutex
        self.lock.acquire()

        self.time_passed_cd = time_passed

        # Release Mutex
        self.lock.release()

        #                                                                          #
        ##                         CRITICAL SECTION END                           ##
        ############################################################################

    def set_time_passed_id(self, time_passed):
        '''set time passed since intermittent dose start'''
        ############################################################################
        ##                         CRITICAL SECTION START                         ##
        #

        # Acquire Mutex
        self.lock.acquire()

        self.time_passed_id = time_passed

        # Release Mutex
        self.lock.release()

        #                                                                          #
        ##                         CRITICAL SECTION END                           ##
        ############################################################################
    def set_time_passed_dl(self, time_passed):
        '''set time passed since delay start'''
        ############################################################################
        ##                         CRITICAL SECTION START                         ##
        #

        # Acquire Mutex
        self.lock.acquire()

        self.time_passed_dl = time_passed

        # Release Mutex
        self.lock.release()

        #                                                                          #
        ##                         CRITICAL SECTION END                           ##
        ############################################################################
    def set_mph(self, max_per_hour_counter):
        '''set max per hour counter'''
        ############################################################################
        ##                         CRITICAL SECTION START                         ##
        #

        # Acquire Mutex
        self.lock.acquire()

        self.mph = max_per_hour_counter

        # Release Mutex
        self.lock.release()

        #                                                                          #
        ##                         CRITICAL SECTION END                           ##
        ############################################################################

    def set_mpi(self, max_per_interval_counter):
        '''set max per interval counter'''
        ############################################################################
        ##                         CRITICAL SECTION START                         ##
        #

        # Acquire Mutex
        self.lock.acquire()

        self.mpi = max_per_interval_counter

        # Release Mutex
        self.lock.release()

        #                                                                          #
        ##                         CRITICAL SECTION END                           ##
        ############################################################################

    def set_status(self, motor_status):
        '''set status'''
        ############################################################################
        ##                         CRITICAL SECTION START                         ##
        #
        # Acquire Mutex
        self.lock.acquire()

        self.status = motor_status

        # Release Mutex
        self.lock.release()

        #                                                                          #
        ##                         CRITICAL SECTION END                           ##
        ############################################################################

    def run(self):
        '''motor run'''
        try:
            # Motor Run
            # print("Motor started at [{}]".format(time.strftime('%H:%M:%S', time.localtime())))
            # print("Motor running ...")
            #
            self.status = MotorStatus.RUNNING
            mph_queue_size = math.ceil(3600/_INTERVAL)
            self.mph_queue = queue.Queue(mph_queue_size)
            while self.status not in [MotorStatus.STOPPED]:
                if self.status == MotorStatus.RUNNING:
                    delta_volume_ab = (self.rate - self.rate_basal) / 3600 * _INTERVAL \
                        if self.rate > self.rate_basal else 0
                    delta_volume = self.rate / 3600 * _INTERVAL
                    self.vinf_ab += delta_volume_ab
                    self.vinf_db += delta_volume_ab
                    self.vinf_cd += delta_volume_ab
                    self.vinf += delta_volume
                    self.vinf_dose += delta_volume
                    self.time_passed += _INTERVAL
                    self.time_passed_ab += _INTERVAL
                    self.time_passed_db += _INTERVAL
                    self.time_passed_cd += _INTERVAL
                    self.time_passed_id += _INTERVAL
                    self.time_passed_dl += _INTERVAL
                    #
                    self.mph += delta_volume
                    self.mph_queue.put(delta_volume)
                    if self.mph_queue.full():
                        self.mph -= self.mph_queue.get()
                    #
                    self.mpi += delta_volume
                    #
                else:
                    # When motor is paused
                    self.time_passed += _INTERVAL
                    self.time_passed_ab += _INTERVAL
                    self.time_passed_db += _INTERVAL
                    self.time_passed_cd += _INTERVAL
                    self.time_passed_id += _INTERVAL
                    #
                    self.mph_queue.put(0)
                    if self.mph_queue.full():
                        self.mph -= self.mph_queue.get()
                    #
                time.sleep(_INTERVAL)
        except KeyboardInterrupt:
            print()
            print('Motor stopped: KeyboardInterrupt')
        except (AttributeError, NameError, OSError):
            print()
            print("{0}: {1}\n".format(sys.exc_info()[0], sys.exc_info()[1]))

    def reset(self):
        self.rate = 0
        self.rate_basal = 0
        self.vinf = 0
        # auto bolus vinf for each ab interval
        self.vinf_ab = 0
        # demand bolus vinf for each demand bolus
        self.vinf_db = 0
        # clinician dose vinf for each clinician dose
        self.vinf_cd = 0
        # intermittent dose vinf for each intermittent dose interval
        self.vinf_dose = 0
        # mph - max per hour counter
        self.mph = 0
        # mpi - max per interval counter
        self.mpi = 0
        self.time_passed = 0
        # time passed since auto bolus start
        self.time_passed_ab = 0
        # time passed since demand bolus start
        self.time_passed_db = 0
        # time passed since clinician dose start
        self.time_passed_cd = 0
        # time passed since intermittent dose start
        self.time_passed_id = 0
        # time passed since delay start
        self.time_passed_dl = 0
        self.status = MotorStatus.STOPPED

def main():
    '''main function'''
    try:
        motor = Motor()
        motor.set_rate(120)
        motor_thread = Thread(target=motor.run, args=())
        motor_thread.start()
        print("Motor started at [{}]".format(time.strftime('%H:%M:%S', time.localtime())))
        print("Motor running ...")
        vtbi = 0.5
        before = time.perf_counter()
        while motor.vinf <= vtbi and motor.status != MotorStatus.STOPPED:
            progress = math.ceil(1000 * motor.vinf / vtbi) / 10
            current = time.perf_counter()
            time_passed = time.strftime('%H:%M:%S', time.gmtime(current - before))
            volume_left = vtbi - motor.vinf if vtbi > motor.vinf else 0
            time_left = time.strftime('%H:%M:%S', time.gmtime(3600 * volume_left / motor.rate))
            print("rate: {:6.2f} ml/h".format(motor.rate), \
                  "vinf: {:6.2f} ml".format(motor.vinf),  \
                  "vtbi: {:6.2f} ml".format(volume_left), \
                  "progress: {:5.1f}%".format(progress), \
                  "time passed: {:4.1f} second".format(motor.time_passed), \
                  "time passed: {} second".format(time_passed), \
                  "time left: {}".format(time_left), sep='  |  ', end='\r')
            time.sleep(_REFRESH_TIMER)
        print()
        print('motor.vinf = {} mL'.format(motor.vinf))
        print('vtbi = {} mL'.format(vtbi))
        motor.set_status(MotorStatus.STOPPED)
        after = time.perf_counter()
        duration = time.strftime('%H:%M:%S', time.gmtime(after - before))
        print('Motor stopped ...')
        print('Motor stopped at [{}]'.format(time.strftime('%H:%M:%S', time.localtime())))
        print('Expected duration = {}'.format(duration))
    except KeyboardInterrupt:
        motor.set_status(MotorStatus.STOPPED)
    except OSError:
        motor.set_status(MotorStatus.STOPPED)
        print()
        print("{0}: {1}\n".format(sys.exc_info()[0], sys.exc_info()[1]))

if __name__ == "__main__":
    main()
    


'''Status Monitor Module'''
import sys
import re
import time
from os import path
import serial
import traceback
from pyZynoUnifiedDrivers.miva.module_miva import Miva
from pyZynoUnifiedDrivers.miva.module_utils import scan_serial_ports
# Time Out (seconds)
_TIMEOUT = 0.05

# Bolus infusion rate (mL/hr)
_BOLUS_RATE = 250

# class PumpEvent(Enum):
    # '''Pump Event'''
    # NONE = 0
    # RUN_INFUSION = 1
    # INFUSION_PAUSED = 2
    # INFUSION_COMPLETED = 3
    # INFUSION_LIMIT_REACHED = 4
    # AUTHENTICATION = 5
    # BOLUS_GRANTED = 6
    # BOLUS_DENIED = 7
    # DOSE_CANCEL = 8
    # OCCLUSION = 9
    # DELAYED_END = 10
    # PUMP_LIMIT_REACHED_ALARM = 11
    # CLEAR_SHIFT_TOTAL = 12
    # BATTERY = 13
    # OPEN_CASSETTE = 14
    # TIMESTAMP = 15
    # POWER_ON = 16
    # POWER_OFF = 17
    # FIRMWARE_ERROR = 18
    # SYSTEM_ERROR = 19
    # BATTERY_FAILURE = 20
    # DEBUG = 21
    # UNATTENDED = 22


class StatusMonitor:
    '''Status Monitor Class'''

    def __init__(self):
        '''__init__'''
        self._is_on = False
        self._is_paused = True
        self.event_log_index_head = '0000'
        self.timestamp_infusion_start = time.time()
        self.local_time_infusion_start = time.localtime()
        #
        self.miva = None
        #
        self.infusion = None

    def start(self, pump):
        '''start'''
        #
        print('Status monitor started at [{}]'\
              .format(time.strftime('%H:%M:%S', time.localtime())))
        print('==')
        #
        self.miva = pump.miva
        miva = self.miva
        self._is_on = True
        self._is_paused = False
        while self._is_on:
            try:
                if self._is_paused:
                    time.sleep(_TIMEOUT)
                    continue
                #
#                 occlusion_sensor = miva.get_occlusion_sensor()
                pump.sensor = miva.get_all_sensor()
#                 pump.sensor['up'] = occlusion_sensor['up']
#                 pump.sensor['down'] = occlusion_sensor['down']
                if pump.sensor == {}:
                    # Stop status monitor
                    self.stop()
                    print('pump.sensor == {}')
                    print('Error: pump failed to respond')
                    print('Stop status monitor...')
                    break
                time.sleep(_TIMEOUT)
            except (serial.serialutil.SerialException, ValueError):
                # Stop status monitor
                self.stop()
                print('Error: pump failed to respond')
                print('Stop status monitor...')
                break
            except (OSError, TypeError, NameError):
                print()
                print("{0}: {1}\n".format(sys.exc_info()[0], sys.exc_info()[1]))
                self.pause()
                raise
            except:
                # Stop status monitor
                self.stop()
                print('Error: pump failed to respond')
                print('Stop status monitor...')
                print()
                print("{0}: {1}\n".format(sys.exc_info()[0], sys.exc_info()[1]))
                traceback.print_exc()
                break

        print('Status monitor stopped at [{}]'\
              .format(time.strftime('%H:%M:%S', time.localtime())))
        print('==')
        # Stop event log monitor
        self.stop()

    def stop(self):
        '''stop'''
        # print('\nStop event log monitor...')
        self._is_paused = True
        self._is_on = False

    def is_on(self):
        '''is on'''
        return self._is_on

    def is_paused(self):
        ''''is paused'''
        is_paused = True
        if self._is_on:
            is_paused = self._is_paused
        else:
            print('Abort: Event log monitor is [OFF]')
            is_paused = None
        return is_paused

    def pause(self):
        ''''pause'''
        if self._is_on:
            self._is_paused = True
        else:
            print('Pause ELM Aborted: Event log monitor is [OFF]')

    def resume(self):
        ''''resume'''
        if self._is_on:
            self._is_paused = False
        else:
            print('Resume ELM Aborted:: Event log monitor is [OFF]')

            
def main(argv):
    '''main function'''
    try:
        current_file = path.basename(__file__)
        if len(argv) == 1:
            # Print the name of current script in Python
            print('Empty argument list. ', end='')
            print('To start, type <{}> followed by the com port. '.format(current_file), end='')
            print('Ex: \'{} COM1\''.format(current_file))
            serial_port_list = scan_serial_ports()
            print('\tAvailable Serial Ports:')
            print('\t\t\t\t', end='')
            for each_serial_port in serial_port_list:
                print('{} '.format(each_serial_port), end='')
        else:
            serial_port = argv[1].upper()
            if re.match(r'com\d+', serial_port.lower().rstrip(' \t\r\n\0')):
                miva = Miva(serial_port)
                status_monitor = StatusMonitor()
                status_monitor.start(miva)
            else:
                print('Error: invalid port \'{}\' '.format(serial_port), end='')
                print('To start, type \'{}\' followed by the com port. '\
                        .format(current_file), end='')
                print('Ex: \'{} COM1\''.format(current_file))
    except KeyboardInterrupt:
        pass
    except serial.serialutil.SerialException:
        print("{0}: {1}\n".format(sys.exc_info()[0], sys.exc_info()[1]))
        traceback.print_exc()


if __name__ == "__main__":
    main(sys.argv)



'''Utility Module'''
import re
import sys
import time
import inspect
import json
import hashlib
import serial
import platform
import struct
from os import path, mkdir, system
from PIL import Image, ImageColor

re_time_str_format = r'(\d\d)(:)(\d\d)'
# Define Flash Structure the same as MIVA pump
FLASH_PAGE_SIZE = 4096
EVENT_LOG_NUMBER_OF_PAGES = 169
EVENT_LOG_SIZE = 16
# 4096 * 169 / 16 = 43264 (A900)
EVENT_LOG_NUMBER_OF_EVENTS = int((FLASH_PAGE_SIZE * EVENT_LOG_NUMBER_OF_PAGES) / EVENT_LOG_SIZE)
# Infusion Data Structure Occupies How Many Event Logs
EVENT_LOGS_PER_INFUSION_DATA_LOG = 16
# Define Regular Expressions
re_scpi_common_cmd = r'^\*[0-9a-zA-Z_]+(\?)?$'
re_scpi_get_cmd = r'^((:[0-9a-zA-Z_]+)+(\?))$'
re_scpi_set_cmd = r'^((:[a-zA-Z]+)+)(\s+)?(\d+)?(\.)?(\d+)?$'
re_get_key_list = r'(:)(key)(:)(list)(\?)((\s+)?(>)(\s+)?([a-z0-9_]+\.txt))?$'
re_get_screenshot = r'(screen(shot)?)((\s+)?(>)(\s+)?(.+\.(png|txt|jpg|json)))?$'
re_compare_screenshot = r'(screen(shot)?)(\s+)?(==)(\s+)?(.+\.(txt|json|jpg|png))?$'
re_query_screenshot_line = r':screen(shot)?:line:[0-8]\?'
re_checksum = r'^(checksum)\s+(status|good|bad)$'
re_clear_key_list = r'(:)(key)(:)(list)(:)(clear)$'
re_compare_parameters = r'^((:[a-zA-Z_]+)+(\?))(\s*)(==|>|>=|<=|<|!=)(\s*)(\S+)(\s+)?(delta(\s*)?==(\s*)?(\d+(\.\d*)?|\.\d+)(%)?)?(\s+)?$'
re_compare_list = r'((:[a-zA-Z]+)+(\?))(\s*)(==|>|>=|<=|<|!=)(\s*)(\[.*\])(\s*)'
re_query_event_timestamp = r'(:)(time(stamp)?)(:)([A-Za-z_]+)(\?)'
re_search_for_event = r'(:)(event(log)?)(:)([A-Za-z_]+)(\?)'
re_query_event_list = r'(:)(event(log)?)(:)(\d+)(\?)'
re_compare_dict_equal = r'^((:[a-zA-Z_]+)+(\?))(\s*)(==)(\s*)(\{.*\})(\s*)$'
re_compare_str = r'^((:[a-zA-Z_]+)+(\?))(\s*)(==|>|>=|<=|<|!=)(\s*)\"(.*)\"(\s+)?$' 
re_compare_str_equal = r'^((:[a-zA-Z_]+)+(\?))(\s*)(==)(\s*)\"(.*)\"(\s+)?$'
re_compare_scpi_return_number = r'^((:[a-zA-Z_]+)+(\?))(\s*)(==|>|>=|<=|<|!=)(\s*)((:[a-zA-Z_]+)+(\?))(\s+)?(delta(\s*)?==(\s*)?(\d+(\.\d*)?|\.\d+)(%)?)?(\s+)?$'
re_query_lockout = r'^(:)(prot(ocol)?|infu(sion)?)(:)(ld|db|sd|ed)(:)(lock(out)?)((:)(min|max|def(ault)?))?(\?)(\s+)?$'
re_qr_code_style = r'^(\d{32})$'
# ^      : matches the start of the string
# (re)   : must contain [re]
# (\s+)? : may contain 1 or more spaces (optional)
# (\d+)  : may contain 0 or more digits
# (\s+)? : may contain 1 or more spaces (optional)
# (>)?   : may contain 0 or 1 [>] (optional)
# (\s+)? : may contain 1 or more spaces (optional)
# ([a-zA-Z0-9_]+\.txt)? 
#        : the file name may contain 1 or more letter and digit 
#          and the underscore [_] and end with [.txt] (optional)
# $      : matches the start of the string 
re_read_event_log = r'^(re)(\s+)?(\d+)?((\s+)?(>)(\s+)?([a-z0-9_]+(\.txt|\.json)))?$'
# ^      : matches the start of the string
# (ref)      : must contain [ref]
# (\s+)?     : must contain 1 or more spaces (optional)
# (\d\d\d\d) : must be 4 digits start index
# (\s+)?     : must contain 1 or more spaces (optional)
# (\d\d\d\d) : must be 4 digits end index
# (>)?       : may contain 0 or 1 [>] (optional)
# (\s+)?     : may contain 1 or more spaces (optional)
# ([a-zA-Z0-9_]+\.txt)? 
#        : the file name may contain 1 or more letter and digit 
#          and the underscore [_] and end with [.txt] (optional)
# $      : matches the start of the string
re_read_event_log_hex = r'^(reh)(\s+)?((0x)?[a-f0-9]+)?(\s+)?((0x)?[a-f0-9]+)?((\s+)?(>)(\s+)?([a-z0-9_]+\.txt))?$'
re_query_help_file = r'^(help|main|\?)(\s+)(.+)(.\s+)?$'
re_ip_address = r'^(?:[0-9]{1,3}\.){3}[0-9]{1,3}$'
re_serial_port = r'(com\d+)|(\/dev\/ttyusb\d+)'

# Define Key Type
_NONE_KEY = '0'
_UP_KEY = '1'
_DOWN_KEY = '2'
_INFO_KEY = '3'
_OK_KEY = '4'
_POWER_KEY = '5'
_RUN_KEY = '6'
_BOLUS_KEY = '7'
_STOP_KEY = '8'
# Define Key Duration
_SHORT_PRESS = '0'
_LONG_PRESS = '1'
_MID_PRESS = '2'

VISA_RESOURCE_CONFIG_FILE_PATH = 'visa_resource_config.json'
PUMP_CONFIG_FILE_PATH = 'pump_config.json'

LCD_MAX_ROWS = 8
LCD_MAX_COL_PIXEL = 128

# Global Variable Definition
_FLASHUTILCL_FILE_PATH = './flash_utilities/FlashUtilCL.exe'.replace('/', '\\')
_MIVA_8_BIT_FIRMWARE_PATH = './firmwares/miva_protective.hex'.replace('/', '\\')
_PYPROGRAMMERCLI_FILE_PATH = './flash_utilities/pyProgrammerCLI.exe'.replace('/', '\\')
_MIVA_32_BIT_FIRMWARE_PATH = './firmwares/miva-32-bit.hex'.replace('/', '\\')
_MFG_32_BIT_FIRMWARE_PATH = './firmwares/mfg-32-bit.hex'.replace('/', '\\')

    
def scan_serial_ports():
    '''Scan Serial Port'''
    ports = []
    if platform.system() == 'Linux':
        ports = ['/dev/ttyUSB%s' % (i) for i in range(0, 256)]
    else:
        ports = ['com%s' % (i) for i in range(1, 256)]
    result = []
    for port in ports:
        try:
            ser = serial.Serial(port)
            ser.close()
            result.append(port)
        except (OSError, serial.SerialException):
            pass
    return result


def wait(seconds):
    '''Wait for number of seconds'''
    while seconds >= 0:
        for each_symbol in ['-', '\\', '|', '/']:
            print('{0}......[{1:.1f} sec]   '.format(each_symbol, seconds), end='\r')
            seconds -= 0.1
            time.sleep(0.1)
            if seconds < 0:
                break
            print('{0}......[{1:.1f} sec]'.format(each_symbol, seconds), end='\r')
            seconds -= 0.1
            time.sleep(0.1)
            if seconds < 0:
                break
            print('{0}......[{1:.1f} sec]'.format(each_symbol, seconds), end='\r')
            seconds -= 0.1
            time.sleep(0.1)
            if seconds < 0:
                break
    # time.sleep(1.5)
    print()
    

def get_line_number():
    '''Returns the current line number in the source code'''
    return inspect.currentframe().f_back.f_lineno


def crc32c(crc, data):
    '''Calculate the CRC32C as Digital Signature
        @param crc The initial CRC
        @param data The data that needs to be calculated, could be hexstring or byte array
        @return CRC of the command
    '''
    crc = (crc ^ 0xffffffff) & 0xffffffff
    i = 0
    poly = 0xedb88320
    # if the [data] is a byte array, convert it to hex string first
    if isinstance(data, bytes):
        data = data.hex().upper()

    while i < len(data):
        crc ^= (int(data[i] + data[i + 1], 16) & 0xff)
        j = 0
        while j < 8:
            crc = (crc >> 1) ^ poly if ((crc & 1) > 0) else crc >> 1
            j += 1
        i += 2
    crc = ~crc
    # return the CRC as a 4-byte integer
    return crc & 0xffffffff


def crc_fill(crc, data, length=-1):
    '''Calculate the CRC32C as Digital Signature and fill the bytes to the length
        @param crc The initial CRC
        @param data The data that needs to be calculated, must be bytes
        @param length The expected length of the output byte array
        @return Bytes with last four bytes been crc, length to be the length parsed in.
         None if length doesn't make sense
    '''
    result = None
    output_hex = False

    if isinstance(data, str):
        output_hex = True
        data = bytes.fromhex(data)
        if length % 2 == 0:
            length = int(length / 2)
        else:
            print("Error: invaild length ", length)
            length = 0

    if length > 5:
        while len(data) < length - 4:
            data += int(0xFF).to_bytes(1, 'little')

        data_hex = data.hex().upper()
        data += crc32c(crc, data_hex).to_bytes(4, 'little')
        if output_hex:
            result = data.hex().upper()
        else:
            result = data

    return result


def byte_fill(data, length, fill_byte=0xFF):
    '''Calculate the CRC32C as Digital Signature and fill the bytes to the length
        @param data The data that needs to be calculated, must be bytes
        @param length The expected length of the output byte array
        @return Bytes array of length. None if length doesn't make sense
    '''
    while len(data) < length:
        data += int(fill_byte).to_bytes(1, 'little')
    return data


def crc_checksum(command, symbol='+'):
    '''Calculate the CRC for Error-Detection Purpose
        @param command R/W+Index+Parameters
        @return CRC of the command
    '''
    summ = 0
    # crc = ''
    for i in range(len(command)):
        # += for nimbus II and CAPS, ^= for nimbus unified H
        if symbol == '+':
            summ += ord(command[i:i + 1])
        elif symbol == '^':
            summ ^= ord(command[i:i + 1])
        else:
            summ += ord(command[i:i + 1])
    crc = hex(summ)[2:].zfill(2)
    return crc[len(crc) - 2: len(crc)]


def ibm_crc16(command):
    '''IBM CRC 16
        Input - input command
        Output - 2-bytes CRC
    '''
    crc = 0xFFFF;
    for k in range(len(command)):
        # Note: Intel architecture sign extends 
        # a char to unsigned short type cast, 
        # so we must mask out the higher byte(s)
        crc ^= command[k] & 0xFF  
        for _ in range(8):
            if crc & 1:
                crc = (crc >> 1) ^ 0xA001
            else:
                crc = (crc >> 1)
    # convert integer to bytes
    crc = crc.to_bytes(2, 'big')
    return crc;


def lrc_checksum(command):
    '''Calculate the LRC (Longitudinal Redundancy Check) checksum
        @param command R/W+Index+Parameters
        @return CRC of the command
    '''
    lrc = 0
    for each_char in command:
        lrc = (lrc + ord(each_char)) & 0xFF
    lrc = ((lrc ^ 0xFF) + 1) & 0xFF
    crc = hex(lrc)[2:].zfill(2)
    print('crc = {}'.format(crc))
    return '0D'


def load_json_file(file_path):
    """Load data from a JSON file"""
    # read JSON file
    with open(file_path, "r", encoding='UTF-8') as file:
        content_json = file.read()
    
    # convert JSON to Python Dict
    config = json.loads(content_json)
    return config


def text_to_file(content, file_path):
    """Save text to a file"""
    dir_name = path.dirname(file_path)
    if dir_name != '':
        if not path.isdir(dir_name):
            mkdir(dir_name)
    if content == None:
        content = ''
    # Save to text file
    with open(file_path, "w+", encoding='UTF-8') as file:
        content = content.replace('\r\n', '\n')
        file.write(content)
    absolute_path = path.abspath(file_path)
    return absolute_path


def json_to_file(config, file_path):
    """Dump data to a JSON file"""
    dir_name = path.dirname(file_path)
    if dir_name != '':
        if not path.isdir(dir_name):
            mkdir(dir_name)   
    if config == None:
        config = {}
    # convert Python Dict to JSON:
    config_json = json.dumps(config, indent=4)
    # Save to JSON file
    with open(file_path, "w+", encoding='UTF-8') as file:
        file.write(config_json)
    absolute_path = path.abspath(file_path)
    return absolute_path


def file_to_list(file_path, ignore_empty_line=True, strip_line=True):
    '''Convert each line in a text file into a list'''
    text_line_list = []
    file = open(file_path, "r")
    for each_line in file:
        if ignore_empty_line:
            if each_line.strip() != '':
                if strip_line:
                    text_line_list.append(each_line.strip())
                else:
                    text_line_list.append(each_line.rstrip())
        else:
            if strip_line:
                text_line_list.append(each_line.strip())
            else:
                text_line_list.append(each_line.rstrip())
    return text_line_list


def list_to_file(content_list, file_path):
    '''Save each element of the list to a text file'''
    dir_name = path.dirname(file_path)
    if dir_name != '':
        if not path.isdir(dir_name):
            mkdir(dir_name)   
            
    file = open(file_path, "w")
    for each_element in content_list: 
        file.write(each_element)
        file.write('\n')
    file.close()
    absolute_path = path.abspath(file_path)
    return absolute_path


def hex_to_float(num_hex):
    '''convert 4-byte hex (little endian) to float 
    ex. '3f800000 -> 1.0' 
    '''
    return_text = struct.unpack('!f', bytes.fromhex(num_hex))[0]    
    return return_text


def float_to_hex(num):
    '''convert float to 4-byte hex (little endian)
    ex. 1.0 -> '3f800000' 
    '''
    return_text = hex(struct.unpack('<I', struct.pack('<f', num))[0])[2:].zfill(8)
    return return_text


def isfloat(num):
    '''Check if number string can be converted to float'''
    try:
        float(num)
        return True
    except ValueError:
        return False


def compare_numbers(num1, num2, operator, delta, delta_is_percentage):
    upper_limit = num2
    lower_limit = num2
    if delta is not None:
        if delta_is_percentage: 
            upper_limit = num2 * (1 + delta / 100)
            lower_limit = num2 * (1 - delta / 100)
        else:
            upper_limit = num2 + delta
            lower_limit = num2 - delta
    
    if operator == '<=':
        if delta is None:
            return (num1 <= num2)
        else:
            return num1 <= upper_limit
    elif operator == '<':
        if delta is None:
            return (num1 < num2)
        else:
            return num1 < upper_limit
    elif operator == '==':
        if delta is None:
            return (num1 == num2)
        else:
            return lower_limit <= num1 <= upper_limit 
    elif operator == '>=':
        if delta is None:
            return (num1 >= num2)
        else:
            return num1 >= lower_limit
    elif operator == '>':
        if delta is None:
            return (num1 > num2)
        else:
            return num1 > lower_limit


def int_to_time_str(time_int, time_format='hh:mm:ss'):
    '''Convert integer to time format [hh:mm] string'''
    time_hour_str = str(int(time_int / 3600)).zfill(2)
    time_minute_str = str(int((time_int % 3600) / 60)).zfill(2)
    time_second_str = str(int((time_int % 3600) % 60)).zfill(2)
    time_str = time_hour_str + ':' + time_minute_str + ':' + time_second_str
    if time_format == 'hh:mm':
        time_str = time_hour_str + ':' + time_minute_str
    return time_str  


def time_str_to_int(time_str):
    '''Convert time format [hh:mm] string to integer'''
    match_result = re.match(re_time_str_format, time_str)
    time_int = int(match_result[1]) * 3600 + int(match_result[3]) * 60
    return time_int  


def hex_to_bitmap(hex_list):
    '''Convert screenshot hex list to bitmap
        ex. the bitmap is a 64x128 list
    '''
    byte_array_list = []
    for i in range(len(hex_list)):
        # Convert each line Hex string to Byte array
        # ex. 'FFFFFFFF' -> b'\xFF\xFF\xFF\xFF'
        byte_array = bytes.fromhex(hex_list[i])
        byte_array_list.append(byte_array)
    
    bit_str_list_list = []
    for i in range(len(byte_array_list)):
        # Convert each line Byte array to bit string list
        bit_str_list = []
        for j in range(len(byte_array_list[i])):
            # Convert Byte to Bit string
            # ex. 255 -> '0b11111111'
            # and ignore the first '0b' in '0b11111111'
            # [bit_str] is the VERTICAL bit string of ONE BYTE
            bit_str = bin(byte_array_list[i][j])[2:].zfill(8)
            bit_str_list.append(bit_str)
        bit_str_list_list.append(bit_str_list)
    
    bitmap_str_list = []
    # print('len(bit_str_list_list) = {}'.format(len(bit_str_list_list)))
    for i in range(len(bit_str_list_list)):
        bit_str_list = bit_str_list_list[i]
        # print('bit_str_list = {}'.format(bit_str_list))
        # print('len(bit_str_list[0]) = {}'.format(len(bit_str_list[0])))
        for j in range(len(bit_str_list[0])):
            bitmap_str_line = ''
            for k in range(len(bit_str_list)):
                # [bitmap_str_line] is the HORIZONTAL bit string WHOLE LINE
                bitmap_str_line += bit_str_list[k][7 - j]
            print(bitmap_str_line)
            bitmap_str_list.append(bitmap_str_line)
    return bitmap_str_list   


def screenshot_hex_to_bitmap(screenshot_hex):
    '''Convert screenshot hex to bitmap
        ex. the bitmap is a 64x128 list
    '''
    
    # convert screenshot_hex to hex_list
    hex_list = []
    for i in range(LCD_MAX_ROWS):
        hex_list.append(screenshot_hex[i * LCD_MAX_COL_PIXEL * 2: (i + 1) * LCD_MAX_COL_PIXEL * 2])
    
    byte_array_list = []
    for i in range(len(hex_list)):
        # Convert each line Hex string to Byte array
        # ex. 'FFFFFFFF' -> b'\xFF\xFF\xFF\xFF'
        byte_array = bytes.fromhex(hex_list[i])
        byte_array_list.append(byte_array)
    
    bit_str_list_list = []
    for i in range(len(byte_array_list)):
        # Convert each line Byte array to bit string list
        bit_str_list = []
        for j in range(len(byte_array_list[i])):
            # Convert Byte to Bit string
            # ex. 255 -> '0b11111111'
            # and ignore the first '0b' in '0b11111111'
            # [bit_str] is the VERTICAL bit string of ONE BYTE
            bit_str = bin(byte_array_list[i][j])[2:].zfill(8)
            bit_str_list.append(bit_str)
        bit_str_list_list.append(bit_str_list)
    
    bitmap_str_list = []
    # print('len(bit_str_list_list) = {}'.format(len(bit_str_list_list)))
    for i in range(len(bit_str_list_list)):
        bit_str_list = bit_str_list_list[i]
        # print('bit_str_list = {}'.format(bit_str_list))
        # print('len(bit_str_list[0]) = {}'.format(len(bit_str_list[0])))
        for j in range(len(bit_str_list[0])):
            bitmap_str_line = ''
            for k in range(len(bit_str_list)):
                # [bitmap_str_line] is the HORIZONTAL bit string WHOLE LINE
                bitmap_str_line += bit_str_list[k][7 - j]
            print(bitmap_str_line)
            bitmap_str_list.append(bitmap_str_line)
    return bitmap_str_list   


def apply_screenshot_masks(screenshot_bitmap, bitmap_masks):
    '''Apply screenshot masks'''
    if bitmap_masks != []:
        for each_mask in bitmap_masks:
            for i in range(each_mask['height']):
                if each_mask['name'] != 'progress_bar':
                    screenshot_bitmap[each_mask['row'] + i] = \
                        screenshot_bitmap[each_mask['row'] + i][:each_mask['column']] \
                        +screenshot_bitmap[each_mask['row']][each_mask['column']] * each_mask['width'] \
                        +screenshot_bitmap[each_mask['row'] + i][each_mask['column'] + each_mask['width']:]
                else:
                    screenshot_bitmap[each_mask['row'] + i] = \
                        screenshot_bitmap[each_mask['row'] + i][:each_mask['column']] \
                        +'0' * each_mask['width'] \
                        +screenshot_bitmap[each_mask['row'] + i][each_mask['column'] + each_mask['width']:]
    
    return screenshot_bitmap


def bitmap_to_image(bitmap_str_list, image_path, bitmap_masks=[]):
    '''Generate photo from Bitmap'''
    # Create directory if not exist
    dir_name = path.dirname(image_path)
    if dir_name != '':
        if not path.isdir(dir_name):
            mkdir(dir_name)
    # Apply masks
    bitmap_str_list = apply_screenshot_masks(bitmap_str_list, bitmap_masks)
    # Create image
    num_line = len(bitmap_str_list)
    num_colume = len(bitmap_str_list[0])
    image = Image.new('1', (num_colume, num_line), color=1)  # create the Image of size 1 pixel 
    for i in range(num_line):
        for j in range(num_colume):
            if bitmap_str_list[i][j] == '1': 
                image.putpixel((j, i), ImageColor.getcolor('black', '1'))
    image = image.resize((num_colume * 1, num_line * 1), resample=0)
    image.save(image_path)
    absolute_path = path.abspath(image_path)
    return absolute_path


def compare_file_equal(file_path, ref_file_path):
    '''Compare two files equal by using MD5'''
    realtime_md5 = hashlib.md5(open(file_path, 'rb').read()).hexdigest()
    reference_md5 = hashlib.md5(open(ref_file_path, 'rb').read()).hexdigest()
    if realtime_md5 == reference_md5:
        return True
    else:
        return False

    
def compare_event_log_equal(event_log, event_log_ref):
    if 'event_type' in event_log and 'event_type' in event_log_ref:
        if event_log['event_type'] == event_log['event_type']:
            return True
        else:
            return False
    else:
        if event_log == event_log_ref:
            return True
        else:
            return False


def validate_qr_code(qr_code, re_qr_code_style):
    # Validate if the QR code is valid
    qr_code = qr_code.strip(' \t\r\n\0')
    if re.match(re_qr_code_style, qr_code.strip(' \t\r\n\0')):
        return True
    return False


def create_pump_config_file(pump_config_file_path):
    # create a [pump_config.json] file
    pump_config = { 
                    "library_path": "",
                    "protocol_selected": "",
                    "rx_path": "",
                    "interval_time": 1,
                    "test_report_path": "./test_report",
                    "event_log": "./event_log",
                    "help_file_path": "./help/miva",
                    "mfg_32_bit_firmware_path": _MFG_32_BIT_FIRMWARE_PATH,
                    "miva_8_bit_firmware_path": _MIVA_8_BIT_FIRMWARE_PATH,
                    "miva_32_bit_firmware_path": _MIVA_32_BIT_FIRMWARE_PATH,
                    "programmer_8_bit_path": _FLASHUTILCL_FILE_PATH,
                    "programmer_32_bit_path": _PYPROGRAMMERCLI_FILE_PATH
                    }
    json_to_file(pump_config, pump_config_file_path)
    return pump_config


def download_mfg_32_bit():
    '''Download MFG 32-bit Firmware'''
    status = False
    #
    # Pump Config
    pump_config = None
    if not path.exists(PUMP_CONFIG_FILE_PATH):
        pump_config = create_pump_config_file(PUMP_CONFIG_FILE_PATH)
    else:
        pump_config = load_json_file(PUMP_CONFIG_FILE_PATH)
    if not 'programmer_32_bit_path' in pump_config:
        pump_config['programmer_32_bit_path'] = _PYPROGRAMMERCLI_FILE_PATH
        print(pump_config)
        json_to_file(pump_config, PUMP_CONFIG_FILE_PATH)
    if not 'mfg_32_bit_firmware_path' in pump_config:
        pump_config['mfg_32_bit_firmware_path'] = _MFG_32_BIT_FIRMWARE_PATH
        json_to_file(pump_config, PUMP_CONFIG_FILE_PATH)
    programmer_32_bit_path = pump_config['programmer_32_bit_path']
    mfg_32_bit_firmware_path = pump_config['mfg_32_bit_firmware_path']
    #
    if not path.exists(programmer_32_bit_path):
        file_basename = path.basename(programmer_32_bit_path)
        print('Fail: 32-bit programmer [{}] NOT exist!'.format(file_basename))
        return
    #
    if not path.exists(mfg_32_bit_firmware_path):
        file_basename = path.basename(mfg_32_bit_firmware_path)
        print('Fail: 32-bit mfg firmware [{}] NOT exist!'.format(file_basename))
        return
    # Download [.hex] with [pyProgrammerCLI.exe]
    # FlashUtilCL Flash [file.hex]
    # [file.hex]: mfg-32-bit.hex
    # .\flash_utilities\pyProgrammerCLI.exe flash .\firmwares\mfg-32-bit.hex> mfg_32_bit_download_log.txt
    command = programmer_32_bit_path + ' flash ' + mfg_32_bit_firmware_path + ' > mfg_32_bit_download_log.txt'
    system(command)
    with open('mfg_32_bit_download_log.txt', "r", encoding='UTF-8') as file:
        file_content = file.read()
        if file_content.strip(' \t\r\n\0') == 'Program done.':
            print("Success: Download [mfg 32-bit firmware] with [pyProgrammerCLI.exe]")
            status = True
        else:
            print(file_content)
    return status


def download_miva_8_bit():
    '''callback function - Download MFG Firmware'''
    status = False
    #
    pump_config = {}
    if not path.exists(PUMP_CONFIG_FILE_PATH):
        pump_config = create_pump_config_file(PUMP_CONFIG_FILE_PATH)
    else:
        pump_config = load_json_file(PUMP_CONFIG_FILE_PATH)
    if not 'programmer_8_bit_path' in pump_config:
        pump_config['programmer_8_bit_path'] = _FLASHUTILCL_FILE_PATH
        json_to_file(pump_config, PUMP_CONFIG_FILE_PATH)
    if not 'miva_8_bit_firmware_path' in pump_config:
        pump_config['miva_8_bit_firmware_path'] = _MIVA_8_BIT_FIRMWARE_PATH
        json_to_file(pump_config, PUMP_CONFIG_FILE_PATH)    
    #
    programmer_8_bit_path = pump_config['programmer_8_bit_path']
    miva_8_bit_firmware_path = pump_config['miva_8_bit_firmware_path']
    
    if not path.exists(programmer_8_bit_path):
        file_basename = path.basename(programmer_8_bit_path)
        print('Fail: 8-bit programmer [{}] NOT exist!'.format(file_basename))
        return
    
    if not path.exists(miva_8_bit_firmware_path):
        file_basename = path.basename(miva_8_bit_firmware_path)
        print('Fail: 8-bit firmware [{}] NOT exist!'.format(file_basename))
        return
    # Download [.hex] with [FlashUtilCL.exe]
    # FlashUtilCL DownloadUSB [-Option Flags] [file.hex] [Serial Number String] [Power On Disconnect] [Debug Interface]
    # [-Option Flags]: N/A
    # [file.hex]: miva-8-bit.hex
    # [Serial Number String]: "" - The serial number of the USB Debug Adapter. 
    #                              An empty string can be used if only one USB 
    #                              Debug Adapter is available.
    # [Power On Disconnect]: 1 - Power will continue to be provided by the USB Debug 
    #                            Adapter to the target device after disconnecting.
    # [Debug Interface]: 1 - C2 Debug Interface (F3XX devices)
    # .\flash_utilities\FlashUtilCL.exe downloadusb .\firmwares\miva_protective.hex "" 1 1 > miva_8_bit_download_log.txt
    command = programmer_8_bit_path + ' downloadusb ' + miva_8_bit_firmware_path + ' \"\" 1 1 > miva_8_bit_download_log.txt'
    system(command)
    # print(command)
    with open('miva_8_bit_download_log.txt', "r", encoding='UTF-8') as file:
        file_content = file.read()
        if file_content.strip(' \t\r\n\0') == '':
            print("Success: Download [miva 8-bit firmware] with [FlashUtilCL.exe]")                
            status = True
        else:
            print(file_content)
            print("Failed: Download [miva 8-bit firmware] with [FlashUtilCL.exe]")
    #
    return status

    
def download_miva_32_bit():
    '''Flash MIVA 32-bit Firmware'''
    status = False
    #
    pump_config = None
    if not path.exists(PUMP_CONFIG_FILE_PATH):
        pump_config = create_pump_config_file(PUMP_CONFIG_FILE_PATH)
    else:
        pump_config = load_json_file(PUMP_CONFIG_FILE_PATH)
    if 'programmer_32_bit_path' not in pump_config:
        pump_config['programmer_32_bit_path'] = _PYPROGRAMMERCLI_FILE_PATH
        json_to_file(pump_config, PUMP_CONFIG_FILE_PATH)
    if not 'miva_32_bit_firmware_path' in pump_config:
        pump_config['miva_32_bit_firmware_path'] = _MIVA_32_BIT_FIRMWARE_PATH
        json_to_file(pump_config, PUMP_CONFIG_FILE_PATH)
    #
    programmer_32_bit_path = pump_config['programmer_32_bit_path']
    miva_32_bit_firmware_path = pump_config['miva_32_bit_firmware_path']
    
    if not path.exists(programmer_32_bit_path):
        file_basename = path.basename(programmer_32_bit_path)
        print('Fail: 32-bit programmer [{}] NOT exist!'.format(file_basename))
        return status
    
    if not path.exists(miva_32_bit_firmware_path):
        file_basename = path.basename(miva_32_bit_firmware_path)
        print('Fail: 32-bit firmware [{}] NOT exist!'.format(file_basename))
        return status
    
    # Download [.hex] with [pyProgrammerCLI.exe]
    # FlashUtilCL Flash [file.hex]
    # [file.hex]: miva-32-bit.hex
    # .\flash_utilities\pyProgrammerCLI.exe flash .\firmwares\miva-32-bit.hex> miva_32_bit_download_log.txt
    command = programmer_32_bit_path + ' flash ' + miva_32_bit_firmware_path + ' > miva_32_bit_download_log.txt'
    system(command)
    # print(command)
    with open('miva_32_bit_download_log.txt', "r", encoding='UTF-8') as file:
        file_content = file.read()
        if file_content.strip(' \t\r\n\0') == 'Program done.':
            print("Success: Download [miva 32-bit firmware] with [pyProgrammerCLI.exe]")
            status = True
        else:
            print(file_content + "\n")
    return status

   
def main(argv):
    '''main function'''
    print(len(argv))
    print('line #: {}'.format(get_line_number()))
    

if __name__ == "__main__":
    main(sys.argv)



'''
Zyno Medical Virtual Instrument Software Architecture (VSIA) python library
Created on May 13, 2022

@author: Yancen Li
'''
import time
import sys
import platform
import serial
from pyZynoUnifiedDrivers.miva.module_miva import Miva
from pyZynoUnifiedDrivers.miva.module_utils import scan_serial_ports

#
_BAUD_RATE = 9600
_BYTE_SIZE = 8
_STOPBITS = 1
_PARITY = serial.PARITY_NONE
_FLOW_CONTROL = 0
_TIMEOUT = 0.025
_WRITE_TIMEOUT = 0.1


class ResourceManager:
    '''
    classdocs
    '''

    def __init__(self):
        '''
        Constructor
        '''
        pass
    
    def list_resources(self):
        '''Scan Serial Port'''
        ports = []
        if platform.system() == 'Linux':
            ports = ['/dev/ttyUSB%s' % (i) for i in range(0, 256)]
        else:
            ports = scan_serial_ports()
        result = []
        for port in ports:
            try:
                ser = serial.Serial(port)
                ser.close()
                result.append(port)
            except (OSError, serial.SerialException):
                pass
        result = tuple(result)
        return result

    def open_resource(self, serial_port):
        '''
        Open Resource
        '''
        return Miva(serial_port)

    def get_pump_serial_port(self):
        com_port = ''
        serial_port_list = scan_serial_ports()
        for each_serial_port in serial_port_list:
            miva = Miva(each_serial_port)
            time.sleep(0.5)
            serial_number = miva.get_pump_sn()
            if serial_number != '':
                com_port = each_serial_port
                miva.close()
                break
            else:
                miva.close()
        return com_port


def main(argv):
    '''main function'''
    rm = ResourceManager()
    serial_port = rm.get_pump_serial_port()
    if serial_port != '':
        miva = rm.open_resource(serial_port)
        # Test [query] Function
        pump_sn = miva.query(':SERIAL?')
        print(pump_sn)
        print('len(pump_sn) = {}'.format(len(pump_sn)))
        # Test [*IDN?] query
        pump_identifier = miva.query('*IDN?')
        print(pump_identifier)
        # Test [close] function of miva class
        miva.close()
    # Test [list_resources] Function
    resources = rm.list_resources()
    print(resources)


if __name__ == "__main__":
    main(sys.argv)



'''nimbus CAPS pump simulator'''
import sys
# import time
from os import path
from threading import Thread 
# from threading import Lock
from pyZynoUnifiedDrivers.miva.module_infusion import Infusion, InfusionStatus, InfusionMonitor
from pyZynoUnifiedDrivers.miva.module_library import Library
from pyZynoUnifiedDrivers.miva.module_utils import load_json_file, json_to_file
from pyZynoUnifiedDrivers.miva.module_utils import PUMP_CONFIG_FILE_PATH
# from module_sendrx import SendRx

# refresh interval (second)
_REFRESH_TIMER = 0.1


class Pump:
    '''pump class'''

    def __init__(self):
        self.sensor = {'up': 0, \
                       'down':0, \
                       'system':0, \
                       'battery_voltage':0, \
                       'motor_running':0}
        self.status_monitor = None
        self.miva = None
        self.connected = False
        self.com_port = ''
        self.serial_number = ''
        self.firmware_version = ''
        self.library_path = ''
        self.library = Library()
        self.infusion = Infusion()
        self.infusion_monitor = InfusionMonitor()
        self.protocol = None
        self.protocol_name = ''
        self.rx_path = ''
        self.infusion.set_status(InfusionStatus.IDLE)
        # Pump configuration file path
        self.pump_config_path = PUMP_CONFIG_FILE_PATH
        # Pump config data structure
        self.pump_config = {"library_path": "", \
							"protocol_selected": "", \
                            "rx_path": "",
                            }

        if path.exists(self.pump_config_path):
            self.pump_config = load_json_file(self.pump_config_path)
            lib_path = self.pump_config["library_path"]
            protocol_name = self.pump_config["protocol_selected"]
            rx_path = self.pump_config["rx_path"]
            if path.exists(lib_path):
                self.library.load(lib_path)
                protocol = self.library.get_protocol(protocol_name)
                if protocol is not None:
                    self.infusion.set_protocol(protocol)
                    self.protocol = protocol
                    self.protocol_name = protocol_name
                if path.exists(rx_path):
                    self.rx_path = rx_path
    
    def start(self):
        '''start console'''
        library = self.library
        infusion = self.infusion
        protocol = self.protocol
        try:
            print('start infusion monitor...')
            print('==')
            self.run_infusion_monitor()
            cmd = ''
            while cmd not in ['exit']:
                cmd = input('>')
                cmd = cmd.lower().strip(' \t\r\n\0')

                if cmd.lower() in ['help', '?']:
                    print(' LL - Load Library')
                    print(' SP - Select Protocol')
                    print(' RP - Run Protocol')

                if cmd.lower() in ['ll', 'load library']:
                    if library.get_library() is not None:
                        print("currently loaded library: ")
                        # Title
                        print("    {0:5s}    |    {1:15s}    ".format('id', 'name'))
                        print("----{0:5s}----|----{1:15s}----".format('-----', '---------------'))
                        # Content
                        print("    {0:5s}    |    {1:15s}"\
                                .format(str(library.get_id()), library.get_name()))
                        lib_path = input('Enter NEW library path: ')
                    else:
                        lib_path = input('    Enter library path: ')
                    if path.exists(lib_path):
                        library.load(lib_path)
                        print("library: ")
                        # Title
                        print("    {0:5s}    |    {1:15s}    ".format('id', 'name'))
                        print("----{0:5s}----|----{1:15s}----".format('-----', '---------------'))
                        # Content
                        print("    {0:5s}    |    {1:15s}"\
                                .format(str(library.get_id()), library.get_name()))
                        #
                        # print("library id = {}   |   ".format(library.get_id()), end='')
                        # print("library name = {}".format(library.get_name()))
                        #
                        self.pump_config["library_path"] = lib_path
                        json_to_file(self.pump_config, self.pump_config_path)
                    else:
                        if lib_path == '':
                            pass
                        else:
                            print('    Invalid library path: \'{}\''.format(lib_path))

                if cmd.lower() in ['sp', 'select protocol']:
                    if library.get_library() is not None:
                        protocols = library.get_protocols()
                        print("library: ")
                        # Title
                        print("    {0:5s}    |    {1:15s}    ".format('id', 'name'))
                        print("----{0:5s}----|----{1:15s}----".format('-----', '---------------'))
                        # Content
                        print("    {:5s}    |    ".format(str(library.get_id())), end='')
                        print("{:15s}".format(library.get_name()))
                        #
                        print("protocol: ")
                        # Title
                        print("    {0:5s}    |    {1:15s}    |    {2:20s}"\
                                .format('id', 'name', 'mode'))
                        print("----{0:5s}----|----{1:15s}----|----{2:20s}"\
                                .format('-----', '---------------', '--------------------'))
                        # Content
                        for each_protocol in protocols:
                            print("    {:5s}    |    ".format(str(each_protocol['id'])), end='')
                            print("{:15s}    |    ".format(each_protocol['content']['name']), end='')
                            print("{:20s}".format(each_protocol['content']['deliveryMode']))
                        print("Total [{}] protocols extracted".format(len(protocols)))
                        #
                        protocol_name = self.protocol_name
                        protocol = library.get_protocol(protocol_name)
                        if protocol is not None:
                            self.infusion.set_protocol(protocol)
                            self.protocol = protocol
                            print("currently selected protocol: ")
                            # Title
                            print("    {0:5s}    |    {1:15s}    |    {2:20s}"\
                                    .format('id', 'name', 'mode'))
                            print("----{0:5s}----|----{1:15s}----|----{2:20s}"\
                                    .format('-----', '---------------', '--------------------'))
                            # Content
                            print("    {0:5s}    |".format(str(protocol['id'])), end='')
                            print("    {0:15s}    |".format(protocol['content']['name']), end='')
                            print("    {0:20s}".format(protocol['content']['deliveryMode']))
                            #
                            print("Extracted Parameters:")
                            if protocol['content']['deliveryMode'] == 'continuousInfusion':
                                infusion.print_cont_parameters()
                            if protocol['content']['deliveryMode'] == 'bolusInfusion':
                                infusion.print_bolus_parameters()
                            if protocol['content']['deliveryMode'] == 'intermittentInfusion':
                                infusion.print_int_parameters()
                            protocol_name = input('Enter NEW protocol name / id: ')
                        else:
                            protocol_name = input('    Enter protocol name / id: ')
                        #
                        if library.get_protocol(protocol_name) is not None:
                            protocol = library.get_protocol(protocol_name)
                            infusion.set_protocol(protocol)
                            print("protocol: ")
                            # Title
                            print("    {0:5s}    |    {1:15s}    |    {2:20s}"\
                                    .format('id', 'name', 'mode'))
                            print("----{0:5s}----|----{1:15s}----|----{2:20s}"\
                                    .format('-----', '---------------', '--------------------'))
                            # Content
                            print("    {0:5s}    |".format(str(protocol['id'])), end='')
                            print("    {0:15s}    |".format(protocol['content']['name']), end='')
                            print("    {0:20s}".format(protocol['content']['deliveryMode']))
                            #
                            print("Extracted Parameters:")
                            if protocol['content']['deliveryMode'] == 'continuousInfusion':
                                infusion.print_cont_parameters()
                            if protocol['content']['deliveryMode'] == 'bolusInfusion':
                                infusion.print_bolus_parameters()
                            if protocol['content']['deliveryMode'] == 'intermittentInfusion':
                                infusion.print_int_parameters()
                            #
                            self.pump_config["protocol_selected"] = protocol_name
                            json_to_file(self.pump_config, self.pump_config_path)
                        else:
                            # invalid protocol name
                            if protocol_name != '':
                                print('    Invalid Protocol: \'{}\''.format(protocol_name))
                    else:
                        # No library loaded
                        print('empty libarary: use <ll> command to load libarary first')
                if cmd.lower() in ['rp', 'run protocol', 'run']:
                    if library.get_library() is not None:
                        if protocol is not None:
                            protocl_name = protocol['content']['name']
                            infusion.set_protocol(protocol)
                            print("protocol: ")
                            # Title
                            print("    {0:5s}    |    {1:15s}    |    {2:20s}"\
                                    .format('id', 'name', 'mode'))
                            print("----{0:5s}----|----{1:15s}----|----{2:20s}"\
                                    .format('-----', '---------------', '--------------------'))
                            # Content
                            print("    {:5s}    |".format(str(protocol['id'])), end='')
                            print("    {:15s}    |".format(protocol['content']['name']), end='')
                            print("    {:20s}".format(protocol['content']['deliveryMode']))
                            #
                            print("Extracted Parameters:")
                            if protocol['content']['deliveryMode'] == 'continuousInfusion':
                                infusion.print_cont_parameters()
                            if protocol['content']['deliveryMode'] == 'bolusInfusion':
                                infusion.print_bolus_parameters()
                            if protocol['content']['deliveryMode'] == 'intermittentInfusion':
                                infusion.print_int_parameters()
                            start_yes_no = input("run protocol - [{}]? (y/n): "\
                                    .format(protocl_name))
                            if start_yes_no.lower().find('y') != -1 or start_yes_no == '':
                                print('infusion start...')
                                self.run_infusion()
                                # time.sleep(_REFRESH_TIMER * 10)
                                # print('infusion stop...')
                            else:
                                print('infusion aborted...')
                        else:
                            print('Empty Protocol: Use <SP> command to select protocol first.')
                    else:
                        print('Empty Libarary: Use <LL> command to load libarary first.')
                #
                if cmd in ['quit', 'status', 'db', 'bolus', 'cd', 'clinician dose', 'stop', 'run', 'pause']:
                    self.infusion_monitor.set_command(cmd)
            # Reset Infusion
            self.infusion.reset()
            # Stop infusion monitor
            self.infusion_monitor.stop()
        except KeyboardInterrupt:
            print()
            print("{0}: {1}\n".format(sys.exc_info()[0], sys.exc_info()[1]))
        except AttributeError:
            print()
            print("{0}: {1}\n".format(sys.exc_info()[0], sys.exc_info()[1]))
        except EOFError:
            # press Ctrl + C
            pass
        except OSError:
            print()
            print("{0}: {1}\n".format(sys.exc_info()[0], sys.exc_info()[1]))
        # Reset Infusion
        self.infusion.reset()
        # Stop infusion monitor
        self.infusion_monitor.stop()

    def run_infusion(self):
        '''Run Infusion'''
        ##############################
        # create a [infusion] thread #
        ##############################
        infusion_thread = Thread(target=self.infusion.run, args=())
        infusion_thread.start()

    def run_infusion_monitor(self):
        '''Run Infusion Monitor'''
        ######################################
        # create a [infusion monitor] thread #
        ######################################
        infusion_monitor_thread = Thread(target=self.infusion_monitor.start, \
                args=([self.infusion]))
        infusion_monitor_thread.start()


def main():
    '''main function'''
    try:
        print('Pump turn on ...')
        pump = Pump()
        pump.start()
        print('Pump turn off ...')
    except KeyboardInterrupt:
        pass


if __name__ == "__main__":
    main()
    
